#    Inside The Data Tunnel - My game for PyWeek 20.
#    Copyright (C) 2015  Jannik Haberbosch
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Modules
# Standard Python modules
import random
import math

# Third party modules
import pygame

# Own modules
from config import config
from classes.Active_Moving_Sprite import Active_Moving_Sprite
from classes.EP_Bar import EP_Bar


# Initialise pygame so that we can use it.
pygame.init()


class Ally( Active_Moving_Sprite ):
	"""The malware class is the base class for all kind of computer-controlled
	malware.  These are the aims of each malware object:
		1)  Get from the "internet" (left side) to the "RAM/CPU" (right side).
		2)  Occasionally, stop and shoot at allys and the players (rotate,
			move and rotate).

	"""
	def __init__( self, _image, _direction, _speed, _turn_speed,
				  _allowed_area, _max_hitpoints, _current_hitpoints,
				  _projectile_speed, _projectile_image, _projectile_damage,
				  _shooting_interval, _kill_score, _additional_dodging_amount,
				  _ep_to_next_level ):
		"""Initialise the object with some predefined values.

		Additional parameters
			_shooting_interval
			int
			Defines the interval for when projectiles will be fired.

			_kill_score
			int
			Defines the score which the player or the allies of the players will
			get once this object has been killed.

		"""
		# Before a call to super, we need to randomly set the position of the
		# Ally object somewhere on the left side where it won't be visible.
		pos_center = (
			random.randint( 1100, 1300 ),
			random.randint( config.ACTIVE_PLAYING_AREA[1], config.ACTIVE_PLAYING_AREA[3] )
		)
		
		super( Ally, self ).__init__( _image, pos_center, _direction,
										 _speed, _turn_speed, _allowed_area,
										 _max_hitpoints, _current_hitpoints,
										 _projectile_speed, _projectile_image,
										 _projectile_damage )

		self.shooting_interval = _shooting_interval * config.FPS
		self.kill_score = _kill_score
		
		self.shooting_interval_counter = 0
		self.phase_interval_counter = 0
		self.phase_interval = 2 * config.FPS

		self.modes =  { "move_forward": True, "stop": False, "attack_malware": False }

		self.move_updown_mode = "up"

		self.ignore_phase_interval_counter = False
		self.additional_dodging_amount = _additional_dodging_amount

		self.ep_to_next_level = _ep_to_next_level
		self.current_ep = 0
		
		# Create the EP bar.
		self.ep_bar = EP_Bar(
			self.rect.width,
			5,
			( self.rect.left, self.rect.top - 13 ),
			config.PURPLE,
			config.DARK_PURPLE,
			self.ep_to_next_level
		)

		# Levels.
		self.current_level = 0

		# Create the level indicator.
		self.create_level_indicator( self.current_level )		

		self.projectiles = pygame.sprite.Group()		


	def change_direction( self, amount ):
		"""Changes the direction to which the malware is moving.

		"""
		self.direction = math.radians


	def has_entered_playing_area( self ):
		"""Returns 'True' once it is visible in the playing area.

		"""
		if self.rect.left > 0:
			return True
		else:
			return False


	def potential_projectiles_detected( self, _other_projectiles ):
		"""This function returns true if at least one potentially dangerous
		projectiles has been detected.

		"""
		for projectile in _other_projectiles:
			if self.rect.top - self.additional_dodging_amount <= projectile.rect.top and \
			   self.rect.bottom + self.additional_dodging_amount >= projectile.rect.bottom:
				return True
			
		return False


	def move_up( self ):
		"""Moves the malware up by x pixels.

		"""
		self.rect.centery -= self.speed


	def move_down( self ):
		"""Moves the malware down by x pixels.

		"""
		self.rect.centery += self.speed


	def move_back( self ):
		"""Moves back.

		"""
		self.rect.left += self.speed


	def decide_move_back( self, _malware ):
		"""Decide if the ally needs to move back.

		"""
		left_positions = []
		for mware in _malware:
			left_positions.append( mware.rect.left + 50 )

		if len( left_positions ) > 0:
			if self.rect.left < max( left_positions ):
				return True
			else:
				return False
		else:
			return False
	

	def update( self, _malware ):
		"""Updates the object and its' behaviour.

		Parameters
			_player_or_allies
			pygame.sprite.Group
			This group contains al players and allies that are still left.
			
		"""
		super( Ally, self ).update( _malware )

		result = False

		# Update the EP bar.
		self.ep_bar.update(
			self.current_ep,
			( self.rect.left, self.rect.top - 13 )
		)

		# Check if we've reached the required amount of EP to proceed to the
		# next level.
		if self.ep_to_next_level <= self.current_ep:
		
			# We have reached the required amount.  Change to the next level.
			# Several properties (such as projectile strength) are now upgraded.
			# Also, tt takes more EP to reach the next level.
			self.ep_to_next_level += 10
			self.current_ep = 0
			self.ep_bar.renew( self.ep_to_next_level )

			self.upgrade_properties(
				self.max_hitpoints + 25,
				( self.projectile_damage[0] + 1, self.projectile_damage[1] + 1 )
			)

			# Change the level.
			self.current_level += 1

			# Change the text of the level indicator so that it shows the correct
			# level.
			new_text = config.LEVEL_INDICATOR["_text"] % str( self.current_level )
			self.level_indicator.change_text( new_text )

		self.level_indicator.reposition( ( self.rect.left, self.rect.top - 23 ) )				
		
		# Here we need to write the algorithm of what should happen each frame.
		# Right now, it is only for testing purposes.

		# The algorithm starts once the malware is visisble:
		if self.is_dead:
			return

		if self.rect.right <= 1024:
			if self.shooting_interval_counter >= self.shooting_interval:
				self.shooting_interval_counter = 0
				super( Ally, self ).fire_projectile()
			else:
				self.shooting_interval_counter += 1

			if self.phase_interval_counter >= self.phase_interval and not self.ignore_phase_interval_counter:
				choice = random.choice( ( "move_forward", "stop", "attack_malware" ) )
				if choice == "move_forward":
					self.modes["move_forward"] = True
					self.modes["stop"] = False
					self.modes["attack_malware"] = False
				elif choice == "stop":
					self.modes["move_forward"] = False
					self.modes["stop"] = True
					self.modes["attack_malware"] = False
				elif choice == "attack_malware":
					self.modes["move_forward"] = False
					self.modes["stop"] = False
					self.modes["attack_malware"] = True

				self.modes["attack_malware"] = True

				self.phase_interval_counter = 0

				if self.move_updown_mode == "up":
					self.move_updown_mode = "down"
				else:
					self.move_updown_mode = "up"
			else:
				self.phase_interval_counter += 1

			if self.modes["move_forward"]:
				if len( _malware ) > 0:
					self.move()

			if self.modes["stop"]:
				pass


			for player in _malware:
				result = self.potential_projectiles_detected( player.projectiles )
				if result:
					if not self.ignore_phase_interval_counter:
						self.ignore_phase_interval_counter = True
					break
				else:
					self.ignore_phase_interval_counter = False

			if not self.ignore_phase_interval_counter: #self.modes["attack_player"]:
				# Choose a player to attack and move so a projectile hits the player.
				l = []
				for x in _malware:
					l.append( x )
				if len( l ) > 0:
					chosen_player = random.choice( l )
					if chosen_player.rect.centery > self.rect.bottom:
						if self.can_move_down():
							self.move_down()
					elif chosen_player.rect.centery < self.rect.top:
						# Move up.
						if self.can_move_up():
							self.move_up()


			# Check for any potential projectiles that may hit the malware.
			# Try to dodge those.  At least try to dodge the one which is nearest.
			if result:
				# We need to dodge!  Check if we can go up.
				if self.can_move_up() or self.can_move_down():
					if self.move_updown_mode == "up":
						# Move up.
						self.move_up()
					else:
						# Move down.
						self.move_down()

			if len( self.projectiles ) > 0:
				self.projectiles.update()

			if self.rect.left > 1024:
				self.kill()

			if self.decide_move_back( _malware ):
				self.move_back()

		else:
			# Otherwise the malware simply moves forward, nothing else.
			self.move()


	def draw( self, window ):	
		super( Ally, self ).draw( window )
		window.blit( self.ep_bar.image, self.ep_bar.rect )
		window.blit( self.level_indicator.image, self.level_indicator.rect )
		

	def reset_position( self ):
		"""Resets the position of an ally once a level has finished.

		"""
		self.rect.center = (
			random.randint( 1100, 1300 ),
			random.randint( config.ACTIVE_PLAYING_AREA[1], config.ACTIVE_PLAYING_AREA[3] )
		)	
