#    Inside The Data Tunnel - My game for PyWeek 20.
#    Copyright (C) 2015  Jannik Haberbosch
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import modules
# Standard modules
import math

# Third party modules
import pygame

# Own modules
from modules import helpers

# In order to use the pygame library, we need to initialise it first.
pygame.init()


class Moving_Sprite( pygame.sprite.Sprite ):
	"""The class 'Moving_Sprite' represents a Sprite that can move around on the
	screen.  It is either moved by a user or by the computer.

	It can move into possibly every single direction (which means it can turn
	aswell).

	And turning means that the image has to be rotated aswell.

	Also, a moving sprite may not move freely around the window, but only within
	a specified area.  Thus, we also need a simply collision-detection
	mechanism.  For simplicity's sake, we are only taking into account a
	collision that may occur in front of the where moving sprite is going.

	"""
	def __init__( self, _image, _pos_center, _direction, _speed, _turn_speed,
				  _allowed_area ):
		"""We use the __init__ function to initialise the object and preload or
		pre-set some of its' variables either by setting them directly, or by
		calling methods.

		Parameters:
			_image
			str
			This is the name of the image of the moving sprite that we want to
			blit onto the window.

			_pos_center
			Python list ( int, int )
			A list with to integer values.  The first one represents the
			centered position of the sprite from the left, the second one
			represents the centered position of the sprite from the top.

			_direction
			integer
			An integer that represents the direction in degrees to which the
			moving sprite is moving.  It can be a value between 0 - 359 degrees.

			_speed
			float
			This defines how many pixels the moving sprite should be moved each
			frame.

			_turn_speed
			integer
			Defines by how many degrees the sprite should be rotated each frame.

			_allowed_area
			pygame.Rect
			This area is a rectangle that defines where the moving sprite is
			allowed to freely move around.  Once the "collsion rectangle" in
			front of the direction of the player is outside this area, the
			moving sprite may not continue moving towards this direction
			anymore, it may only turn until the "collsion rectangle" is once
			again contained in the allowed area.

		"""
		# We need to initialise the base-class first so that we can use its'
		# methods and variables.
		pygame.sprite.Sprite.__init__( self )

		self.image = helpers.load_image( _image )
		self.rect = self.image.get_rect( center = _pos_center )

		# We need a copy (not a reference!) of the image that has been passed to
		# this class.  This is because we may need to rotate the image  when the
		# moving sprite rotates.  Using a rotated image for additional rotation
		# would return in a loss of quality each time the rotated image is
		# rotated again.
		self.original_image = self.image.copy()

		# Although the direction is passed in in degrees( because that is easier
		# to understand for me ), Python prefers to use radians ( I think ),
		# because almost every mathematical function uses radians.
		# This is why I am going to convert the value in degrees to radians.
		self.direction = math.radians( _direction )

		self.speed = _speed
		# We need to calculate the speed along the x- and the y-axis so that we
		# can move the moving sprite on our window correctly.
		# For this task, I am going to make use of a function, because this is
		# a functionality we are going to need quite often ( as soon as the
		# direction changes ).
		self.x_axis_speed = 0
		self.y_axis_speed = 0
		self.calc_x_axis_speed()
		self.calc_y_axis_speed()

		self.turn_speed = _turn_speed

		self.allowed_area = _allowed_area

		# The "collision rectangle" is a special pygame.Rect object that is
		# always located in front of the moving sprites' direction.  It is used
		# to determine whether there is a collsion with one of the walls of the
		# 'allowed_area' or not.
		self.collision_rectangle = pygame.Rect( 0, 0, self.image.get_size()[0],
											    self.image.get_size()[1] )
		# And we need to position our collision rectangle.
		self.update_collision_rectangle_position()

		# We need another variable to help tell the classes that inherit from
		# this class whether the sprite is allowed to move or not.
		self.is_allowed_to_move = True


	def calc_x_axis_speed( self ):
		"""This function calculates the speed along the x axis for our moving
		sprite.

		"""
		self.x_axis_speed = math.sin( self.direction ) * self.speed


	def calc_y_axis_speed( self ):
		"""This function calculates the speed along the y axis for our moving
		sprite.

		"""
		self.y_axis_speed = math.cos( self.direction ) * self.speed


	def get_x_axis_speed( self ):
		"""This function returns the speed of the moving sprite along the
		x-axis.

		"""
		return self.x_axis_speed


	def get_y_axis_speed( self ):
		"""This function returns the speed of the moving sprite along the
		y-axis.

		"""
		return self.y_axis_speed


	def get_speed( self ):
		"""This function returns the speed of the moving sprite.

		"""
		return self.speed


	def get_direction( self ):
		"""This function returns the direction of the moving sprite in degrees.

		"""
		return math.degrees( self.direction )


	def move( self ):
		"""This function moves the sprite.

		"""
		self.rect.move_ip( self.x_axis_speed, self.y_axis_speed )


	def turn_right( self ):
		"""This function changes the direction a moving sprite is facing towards
		the right.

		"""
		self.direction = self.get_direction() - self.turn_speed
		if self.direction < 0:
			self.direction = 360 - self.turn_speed

		self.direction = math.radians( self.direction )


	def turn_left( self ):
		"""This function changes the direction a moving sprite is facing towards
		the left.

		"""
		self.direction = self.get_direction() + self.turn_speed
		if self.direction > 360:
			self.direction = self.turn_speed

		self.direction = math.radians( self.direction )
		
	
	def rotate_image( self ):
		"""This function rotates the image and sets the current position once
		again.

		"""
		position_center = self.rect.center
		
		self.image = pygame.transform.rotate(
			self.original_image,
			self.get_direction()
		)

		self.rect = self.image.get_rect( center = position_center )


	def is_outside_allowed_area( self ):
		"""This function returns 'True' if the moving sprite is outside the
		allowed area, otherwise 'False'.

		"""
		if not self.allowed_area.contains( self.collision_rectangle ):
			return True
		else:
			return False


	def update_collision_rectangle_position( self ):
		"""This function updates the position (sets the position) of the
		collision rectangle.

		"""
		self.collision_rectangle.center = (
			self.rect.centerx + math.sin( self.direction ) * self.original_image.get_width(),
			self.rect.centery + math.cos( self.direction ) * self.original_image.get_height()
		)


	def get_position_center( self ):
		"""This function returns the center position of the sprite as a tuple.

		"""
		return self.rect.center


	def set_position_center( self, _new_center ):
		"""This function setes the center position of the sprite.

		"""
		self.rect.center = _new_center


	def update( self ):
		"""This function updates the sprite or its' behaviour.

		"""
		self.update_collision_rectangle_position()
		
		if self.is_outside_allowed_area():
			self.is_allowed_to_move = False
			
		elif not self.is_allowed_to_move:
			self.is_allowed_to_move = True
			

	def can_move( self ):
		"""This function returns 'True' if the moving sprite is allowed to move
		Note that 'move' is not the same as 'rotate'.

		This function will most likely return 'False' in case of a collision.

		"""
		return self.is_allowed_to_move


	def get_collision_rectangle( self ):
		"""Returns the collision rectangle.

		"""
		return self.collision_rectangle
